/*----------------------------------------------------------------------
  Linear algebra routines for vectors
  File:    linmath.h
  Purpose: 
  Author:  Tomas Mller
           Department of Computer Engineering
           Chalmers University of Technology   
  History: July 1999 (started)

----------------------------------------------------------------------*/

#ifndef LINMATH_H
#define LINMATH_H

#include <math.h>
#include <stdio.h>

typedef float Vec2f[2];
typedef double Vec2d[2];

typedef float Vec3f[3]; 
typedef double Vec3d[3];

typedef float Vec4f[4];
typedef double Vec4d[4];

typedef float Mat33f[3][3];
typedef double Mat33d[3][3];

typedef float Mat44f[4][4];
typedef double Mat44d[4][4];

#define X 0
#define Y 1
#define Z 2
#define W 3

#define R 0
#define G 1
#define B 2
#define A 3 /* alpha */

/* V3SET(a,b) is a=b */
#define V3SET(a,b) (a)[X]=(b)[X]; (a)[Y]=(b)[Y]; (a)[Z]=(b)[Z];

#define V3SET3(a,x,y,z) (a)[X]=x; (a)[Y]=y; (a)[Z]=z;


/* V4SET(a,b) is a=b */
#define V4SET(a,b) (a)[X]=(b)[X]; (a)[Y]=(b)[Y]; (a)[Z]=(b)[Z]; (a)[W]=(b)[W];

#define V4SET4(a,x,y,z,w) (a)[X]=x; (a)[Y]=y; (a)[Z]=z; (a)[W]=w;

/* V3ADD(d,u,v) is d=u+v */
#define V3ADD(d,u,v) (d)[X]=(u)[X]+(v)[X]; (d)[Y]=(u)[Y]+(v)[Y]; (d)[Z]=(u)[Z]+(v)[Z];

/* V3SUB(d,u,v) is d=u-v */
#define V3SUB(d,u,v) (d)[X]=(u)[X]-(v)[X]; (d)[Y]=(u)[Y]-(v)[Y]; (d)[Z]=(u)[Z]-(v)[Z];

/* V3SCALE(d,u,s) is d=s*u (s=scalar) */
#define V3SCALE(d,u,s) (d)[X]=(s)*(u)[X]; (d)[Y]=(s)*(u)[Y]; (d)[Z]=(s)*(u)[Z];

/* V3DOT(a,b) is the dot product */
#define V3DOT(a,b) ((a)[X]*(b)[X]+(a)[Y]*(b)[Y]+(a)[Z]*(b)[Z])
#define V4DOT(a,b) ((a)[X]*(b)[X]+(a)[Y]*(b)[Y]+(a)[Z]*(b)[Z]+(a)[W]*(b)[W])

/* V3DIST(a,b): d is the distance between point a and point b */
#define V3DIST(d,a,b) \
{	Vec3f v; \
	V3SUB(v,b,a); \
	d=sqrt(V3DOT((v),(v))); \
}

/* V3DIST(a,b): d is the distance between point a and point b */
#define V3DISTD(d,a,b) \
{	Vec3d v; \
	V3SUB(v,b,a); \
	d=sqrt(V3DOT((v),(v))); \
}

/* V3LERP is dest=(1-t)*a + t*b */
#define V3LERP(dest,t,a,b) \
  (dest)[X]=(1.0-(t))*(a)[X]+(t)*(b)[X]; \
  (dest)[Y]=(1.0-(t))*(a)[Y]+(t)*(b)[Y]; \
  (dest)[Z]=(1.0-(t))*(a)[Z]+(t)*(b)[Z]; 

/* V3CROSS is the cross product: d= u x v */
#define V3CROSS(d,u,v)       \
   (d)[X]=(u)[Y]*(v)[Z]-(u)[Z]*(v)[Y]; \
   (d)[Y]=(u)[Z]*(v)[X]-(u)[X]*(v)[Z]; \
   (d)[Z]=(u)[X]*(v)[Y]-(u)[Y]*(v)[X]; 

#define V3NORMALIZEF(v) \
{  float s=(float)sqrt((float)V3DOT((v),(v)));         \
   if(s>0.00000001)                       \
   {  float is=1.0f/s;                    \
      (v)[X]*=is; (v)[Y]*=is; (v)[Z]*=is; \
   }                                      \
}

#define V3NORMALIZED(v) \
{  double s=sqrt(V3DOT((v),(v)));          \
   if(s>0.00000000001)                    \
   {  double is=1.0/s;                   \
      (v)[X]*=is; (v)[Y]*=is; (v)[Z]*=is; \
   }                                      \
}

/* V3NORMAL is n=(v1-v0)x(v2-v0) */
#define V3NORMALF(n,v0,v1,v2) \
{ Vec3f v10,v20;      \
  V3SUB(v10,v1,v0);   \
  V3SUB(v20,v2,v0);   \
  V3CROSS(n,v10,v20); \
}

#define V3PRINT(str,v) printf("%s: (%2.4f,%2.4f,%2.4f)\n",str,\
			      (float)((v)[X]),(float)((v)[Y]),(float)((v)[Z])); 

#define V3PRINT2(v) printf(" %g %g %g ", \
			      (float)((v)[X]),(float)((v)[Y]),(float)((v)[Z])); 

/* create an Euler rotation matrix: R_z(roll)*R_x(pitch)*R_y(head)=euler */
#define M33EULER(euler,head,pitch,roll) \
{  double ch,sh,cp,sp,cr,sr; \
   ch=cos(head);  \
   sh=sin(head);  \
   cp=cos(pitch); \
   sp=sin(pitch); \
   cr=cos(roll);  \
   sr=sin(roll);  \
   (euler)[X][X]=cr*ch-sr*sp*sh; \
   (euler)[X][Y]=-sr*cp; \
   (euler)[X][Z]=cr*sh+sr*sp*ch; \
   (euler)[Y][X]=sr*ch+cr*sp*sh; \
   (euler)[Y][Y]=cr*cp; \
   (euler)[Y][Z]=sr*sh-cr*sp*ch; \
   (euler)[Z][X]=-cp*sh; \
   (euler)[Z][Y]=sp; \
   (euler)[Z][Z]=cp*ch; \
}

/* make a matrix of three vectors: dest=(col0 col1 col2) */
#define M33COLUMNS(dest,col0,col1,col2) \
   (dest)[X][X]=(col0)[X]; \
   (dest)[Y][X]=(col0)[Y]; \
   (dest)[Z][X]=(col0)[Z]; \
   (dest)[X][Y]=(col1)[X]; \
   (dest)[Y][Y]=(col1)[Y]; \
   (dest)[Z][Y]=(col1)[Z]; \
   (dest)[X][Z]=(col2)[X]; \
   (dest)[Y][Z]=(col2)[Y]; \
   (dest)[Z][Z]=(col2)[Z]; 

/* destvec=matrix*vec */
#define M33MULTVEC(destvec,matrix,vec) \
  (destvec)[X]=(matrix)[X][X]*(vec)[X]+(matrix)[X][Y]*(vec)[Y]+(matrix)[X][Z]*(vec)[Z];\
  (destvec)[Y]=(matrix)[Y][X]*(vec)[X]+(matrix)[Y][Y]*(vec)[Y]+(matrix)[Y][Z]*(vec)[Z];\
  (destvec)[Z]=(matrix)[Z][X]*(vec)[X]+(matrix)[Z][Y]*(vec)[Y]+(matrix)[Z][Z]*(vec)[Z];

/* destvec=matrix*vec */
#define M43MULTVEC(destvec,matrix,vec) \
  (destvec)[X]=(matrix)[X][X]*(vec)[X]+(matrix)[X][Y]*(vec)[Y]+(matrix)[X][Z]*(vec)[Z]+(matrix)[X][W]*(vec)[W];\
  (destvec)[Y]=(matrix)[Y][X]*(vec)[X]+(matrix)[Y][Y]*(vec)[Y]+(matrix)[Y][Z]*(vec)[Z]+(matrix)[Y][W]*(vec)[W];\
  (destvec)[Z]=(matrix)[Z][X]*(vec)[X]+(matrix)[Z][Y]*(vec)[Y]+(matrix)[Z][Z]*(vec)[Z]+(matrix)[Z][W]*(vec)[W];

/* destvec=matrix*vec */
#define M44MULTVEC(destvec,matrix,vec) \
  (destvec)[X]=(matrix)[X][X]*(vec)[X]+(matrix)[X][Y]*(vec)[Y]+(matrix)[X][Z]*(vec)[Z]+(matrix)[X][W]*(vec)[W];\
  (destvec)[Y]=(matrix)[Y][X]*(vec)[X]+(matrix)[Y][Y]*(vec)[Y]+(matrix)[Y][Z]*(vec)[Z]+(matrix)[Y][W]*(vec)[W];\
  (destvec)[Z]=(matrix)[Z][X]*(vec)[X]+(matrix)[Z][Y]*(vec)[Y]+(matrix)[Z][Z]*(vec)[Z]+(matrix)[Z][W]*(vec)[W];\
  (destvec)[W]=(matrix)[W][X]*(vec)[W]+(matrix)[W][Y]*(vec)[Y]+(matrix)[W][Z]*(vec)[Z]+(matrix)[W][W]*(vec)[W];

/* dest=m*n */
#define M33MULTMM(dest,m,n) \
  (dest)[X][X]=(m)[X][X]*(n)[X][X]+(m)[X][Y]*(n)[Y][X]+(m)[X][Z]*(n)[Z][X]; \
  (dest)[X][Y]=(m)[X][X]*(n)[X][Y]+(m)[X][Y]*(n)[Y][Y]+(m)[X][Z]*(n)[Z][Y]; \
  (dest)[X][Z]=(m)[X][X]*(n)[X][Z]+(m)[X][Y]*(n)[Y][Z]+(m)[X][Z]*(n)[Z][Z]; \
  (dest)[Y][X]=(m)[Y][X]*(n)[X][X]+(m)[Y][Y]*(n)[Y][X]+(m)[Y][Z]*(n)[Z][X]; \
  (dest)[Y][Y]=(m)[Y][X]*(n)[X][Y]+(m)[Y][Y]*(n)[Y][Y]+(m)[Y][Z]*(n)[Z][Y]; \
  (dest)[Y][Z]=(m)[Y][X]*(n)[X][Z]+(m)[Y][Y]*(n)[Y][Z]+(m)[Y][Z]*(n)[Z][Z]; \
  (dest)[Z][X]=(m)[Z][X]*(n)[X][X]+(m)[Z][Y]*(n)[Y][X]+(m)[Z][Z]*(n)[Z][X]; \
  (dest)[Z][Y]=(m)[Z][X]*(n)[X][Y]+(m)[Z][Y]*(n)[Y][Y]+(m)[Z][Z]*(n)[Z][Y]; \
  (dest)[Z][Z]=(m)[Z][X]*(n)[X][Z]+(m)[Z][Y]*(n)[Y][Z]+(m)[Z][Z]*(n)[Z][Z]; 

/* d=transpose(m) */
#define M44TRANSPOSE(d,m) \
 (d)[X][X]=(m)[X][X]; (d)[X][Y]=(m)[Y][X]; (d)[X][Z]=(m)[Z][X]; (d)[X][W]=(m)[W][X]; \
 (d)[Y][X]=(m)[X][Y]; (d)[Y][Y]=(m)[Y][Y]; (d)[Y][Z]=(m)[Z][Y]; (d)[Y][W]=(m)[W][Y]; \
 (d)[Z][X]=(m)[X][Z]; (d)[Z][Y]=(m)[Y][Z]; (d)[Z][Z]=(m)[Z][Z]; (d)[Z][W]=(m)[W][Z]; \
 (d)[W][X]=(m)[X][W]; (d)[W][Y]=(m)[Y][W]; (d)[W][Z]=(m)[Z][W]; (d)[W][W]=(m)[W][W];

#define M33PRINT(m) \
  printf("(%2.4f %2.4f %2.4f)\n",m[X][X],m[X][Y],m[X][Z]); \
  printf("(%2.4f %2.4f %2.4f)\n",m[Y][X],m[Y][Y],m[Y][Z]); \
  printf("(%2.4f %2.4f %2.4f)\n\n",m[Z][X],m[Z][Y],m[Z][Z]); 

#define M44PRINT(m) \
  printf("(%2.4f %2.4f %2.4f %2.4f)\n",m[X][X],m[X][Y],m[X][Z],m[X][W]); \
  printf("(%2.4f %2.4f %2.4f %2.4f)\n",m[Y][X],m[Y][Y],m[Y][Z],m[Y][W]); \
  printf("(%2.4f %2.4f %2.4f %2.4f)\n",m[Z][X],m[Z][Y],m[Z][Z],m[Z][W]); \
  printf("(%2.4f %2.4f %2.4f %2.4f)\n",m[W][X],m[W][Y],m[W][Z],m[W][W]); 

/* get quaternion from matrix m, store result in Vec4{fd} q */
/* I ought to put this in a .c file /TM */
static void M33Quaternion(Mat33f m,Vec4d q)
{
    double trace = m[0][0]+m[1][1]+m[2][2]; 
    double root;        
    if(trace>0.0)       
    {                   
       root = sqrt(trace+1.0);     
       q[3] = 0.5*root;               
       root = 0.5/root;            
       q[0] = (m[1][2]-m[2][1])*root; 
       q[1] = (m[2][0]-m[0][2])*root; 
       q[2] = (m[0][1]-m[1][0])*root; 
    }        
    else     
    {        
       static int next[3] = { 1, 2, 0 }; 
       int j,k,i = 0; 
       if(m[1][1] > m[0][0]) 
	  i = 1; 
       if(m[2][2] > m[i][i]) 
	  i = 2; 
       j = next[i]; 
       k = next[j]; 
       root = sqrt(m[i][i]-m[j][j]-m[k][k]+1.0); 
       q[i] = 0.5*root; 
       root = 0.5/root; 
       q[j] = (m[i][j]+m[j][i])*root; 
       q[k] = (m[i][k]+m[k][i])*root; 
       q[3] = (m[j][k]-m[k][j])*root; 
    } 
}

// UA ADD
/* I ought to put this in a .c file /TM */
static void M33RotationAxis(Vec3f axis,float radians,Mat33f mtx)
/* creates rotation matrix, rotates "radians" around "axis", axis must be normalized */
{
   float sinalpha,cosalpha;
   float q[4],Nq,s,xs,ys,zs,wx,wy,wz,xx,xy,xz,yy,yz,zz;
   /* from Ken Shoemake's gem: uses a quaternion to construct the matrix*/
   radians*=0.5;
   sinalpha=(float)sin(radians);
   cosalpha=(float)cos(radians);

   /* create the quaternion */
   q[0]=axis[0]*sinalpha;
   q[1]=axis[1]*sinalpha;
   q[2]=axis[2]*sinalpha;
   q[3]=cosalpha;

   Nq=q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3];
   s=(Nq > 0.0f) ? (2.0f / Nq) : 0.0f;
   xs=q[0]*s;  ys=q[1]*s;  zs=q[2]*s;
   wx=q[3]*xs; wy=q[3]*ys; wz=q[3]*zs;
   xx=q[0]*xs; xy=q[0]*ys; xz=q[0]*zs;
   yy=q[1]*ys; yz=q[1]*zs; zz=q[2]*zs;
   
   mtx[0][0]=1.0f-(yy + zz); mtx[0][1]=xy - wz;     mtx[0][2]=xz + wy; 
   mtx[1][0]=xy + wz;        mtx[1][1]=1.0f-(xx+zz); mtx[1][2]=yz - wx;  
   mtx[2][0]=xz - wy;        mtx[2][1]=yz + wx;      mtx[2][2]=1.0f-(xx+yy);
}

#endif
