
#ifndef _include_img_Image_hpp_
#define _include_img_Image_hpp_ 


// includes
#include <iostream>

namespace bleifrei {
	namespace platform {
		
		// defs
		typedef unsigned  char	byte;
		typedef unsigned  short	word;
		typedef unsigned  long  dword;
		typedef signed    char  sbyte;
		typedef signed    short sword;
		typedef	signed    long  sdword;
		
	}
}

namespace bleifrei {
	namespace img {

		class stream_exception {};
		class out_of_memory_exception {};
		class file_format_exception {};

		enum ImageCompression
		{
			ImageCompressionRaw,
			ImageCompressionRLE,
			ImageCompressionMax
		};

		enum ImageFormat
		{
			ImageFormatA8,
			ImageFormatL8,
            ImageFormatR5G5B5A1,
            ImageFormatB5G5R5A1,
			ImageFormatR5G6B5,
			ImageFormatB5G6R5,
			ImageFormatR8G8B8,
			ImageFormatB8G8R8,
			ImageFormatR8G8B8A8,
			ImageFormatB8G8R8A8,
			ImageFormatMax
		};

		/** An image class.
		 * Copyright 2003 (c) \n
		 * A class that can hold an image of a specific width, height and bit depth. The
		 * image can be loaded from/saved to file in both compressed and uncompressed form.
		 * The method copy can be used to copy the image data from another image and if
		 * necessary convert it to a new bit depth. \n
		 * The methods getPixel and putPixel allows for easy manipulation of specific pixels
		 * inside the image
		 * @author Ulf Borgenstam
		 */
		class Image
		{
		public:
			// constructors
			Image(void);
			Image(std::istream &is);
			Image(ImageFormat format);
			Image(int width, int height, ImageFormat format);
			Image(std::istream &is, int width, int height, ImageFormat format);

			// destructor
			virtual ~Image(void);

			// public methods
			void create(int width, int height, ImageFormat format);
			void destroy(void);
			void flip(void);
			void putPixel(int x, int y, platform::byte r, platform::byte g, platform::byte b, platform::byte alpha = 0);
			void getPixel(int x, int y, platform::byte &r, platform::byte &g, platform::byte &b, platform::byte &alpha);
			void copy(Image &image, ImageFormat format);

			// stream operations
			void decodeRLE(std::istream &is, int numbytes);
			void encodeRLE(std::ostream &os);
			void save(std::ostream &os, ImageCompression compression = ImageCompressionRaw);
			void load(std::istream &is);

			// public constant methods
			int getWidth(void) const											{ return m_width;						}
			int getHeight(void) const											{ return m_height;						}
			int getPixelWidth(void) const										{ return m_pixelwidth;					}
			int getBytesPerLine(void) const										{ return m_bytesPerLine;				}
			ImageFormat getFormat(void) const									{ return m_format;						}
			platform::byte *getData(void)										{ return m_imageData;					}

		protected:

			void convertBGRtoRGB32();
			void convertBGRtoRGB24();
			void convertBGRtoRGB();

			// attributes
			platform::byte *m_imageData;
			int m_width, m_height;
			int m_bytesPerLine;
			int m_pixelwidth;
			ImageFormat m_format;
		};

		inline ImageFormat
		bitDepthToImageFormat(int bpp)
		{
			switch (bpp)
			{
			case 8 :
				return ImageFormatA8;
			case 15 :
				return ImageFormatR5G5B5A1;
			case 16 :
				return ImageFormatR5G6B5;
			case 24 :
				return ImageFormatR8G8B8;
			case 32 :
				return ImageFormatR8G8B8A8;
			default :
				return ImageFormatMax;
			}
		}

		inline ImageFormat
		bitDepthToImageFormatRev(int bpp)
		{
			switch (bpp)
			{
			case 8 :
				return ImageFormatA8;
			case 15 :
				return ImageFormatB5G5R5A1;
			case 16 :
				return ImageFormatB5G6R5;
			case 24 :
				return ImageFormatB8G8R8;
			case 32 :
				return ImageFormatB8G8R8A8;
			default :
				return ImageFormatMax;
			}
		}

		inline int
		imageFormatToBitDepth(ImageFormat format)
		{
			switch (format)
			{
			case ImageFormatA8 :
			case ImageFormatL8 :
				return 8;
			case ImageFormatR5G5B5A1 :
			case ImageFormatB5G5R5A1 :
				return 15;
			case ImageFormatR5G6B5 :
			case ImageFormatB5G6R5 :
				return 15;
			case ImageFormatR8G8B8 :
			case ImageFormatB8G8R8 :
				return 24;
			case ImageFormatR8G8B8A8 :
			case ImageFormatB8G8R8A8 :
				return 32;
			default:
				return 0;
			}
		}

	}
}

#endif //_include_bleifrei_img_Image_hpp_
