#ifndef BOUNDINGBOX_h
#define BOUNDINGBOX_h

#include "vecmath.h"
#include <float.h>

class BoundingBox
{
public:
	BoundingBox(void) 
	{
		m_minCorner = Vec3f(FLT_MAX, FLT_MAX, FLT_MAX);
		m_maxCorner = Vec3f(-FLT_MAX, -FLT_MAX, -FLT_MAX);
	}

	/**
	 * A constructor
	 * @param minCorner The corner of the box with the smallest coordinates
	 * @param maxCorner The corner of the box with the greatest coordinates
	 */
	BoundingBox(const Vec3f &minCorner, const Vec3f &maxCorner) : 
		m_minCorner(minCorner), m_maxCorner(maxCorner) {}

//OPERATORS
	/**
	 * The assignment operator.
	 * @param box The BoundingBox to assign.
	 */
	BoundingBox &operator=(const BoundingBox &box);

//ACCESS
	/**
	 * Sets the minimum corner.
	 * @param minCorner the minimum corner
	 */
	void setMinCorner(const Vec3f &minCorner);

	/**
	 * Sets the maximum corner.
	 * @param maxCorner the maximum corner
	 */
	void setMaxCorner(const Vec3f &maxCorner);

	/**
	 * Gets the minimum corner.
	 * @return the minimum corner
	 */
	Vec3f &getMinCorner(void);
	const Vec3f &getMinCorner(void) const;

	/**
	 * Gets the maximum corner.
	 * @return the maximum corner
	 */
	Vec3f &getMaxCorner(void);
	const Vec3f &getMaxCorner(void) const;

//OPERATIONS
	inline void operator+=(const Vec3f& p) 
	{
		for (int i=0; i<3; ++i)
		{
			if (p[i] < m_minCorner[i])
				m_minCorner[i] = p[i];
			if (p[i] > m_maxCorner[i])
				m_maxCorner[i] = p[i];
		}
	}

	inline void operator+=(const BoundingBox &bb) 
	{
		const Vec3f& mx = bb.getMaxCorner();
		const Vec3f& mn = bb.getMinCorner();
		for (int i = 0; i < 3; i++) {
			if (mn[i] < m_minCorner[i])
				m_minCorner[i] = mn[i];
			if (mx[i] > m_maxCorner[i])
				m_maxCorner[i] = mx[i];
		}
	}

	/**
	 * Checks if this position is inside the box
	 * @param point the point to be tested
	 * @return true if point is inside box
	 */
	bool isInside(const Vec3f &point) const;

	Vec3f centerPoint(void) const;
	
	Vec3f halfVector(void) const;

	Vec3f size(void) const;

	void grow(float minSize, float percentGrow);

private:
	Vec3f m_minCorner;
	Vec3f m_maxCorner;
};

inline BoundingBox &BoundingBox::operator=(const BoundingBox &box) {
	m_minCorner = box.m_minCorner;
	m_maxCorner = box.m_maxCorner;
	return *this;
}

inline void BoundingBox::setMinCorner(const Vec3f &minCorner) {
	m_minCorner = minCorner;
}

inline void BoundingBox::setMaxCorner(const Vec3f &maxCorner) {
	m_maxCorner = maxCorner;
}

inline Vec3f &BoundingBox::getMinCorner(void) {
	return m_minCorner;
}

inline const Vec3f &BoundingBox::getMinCorner(void) const {
	return m_minCorner;
}

inline Vec3f &BoundingBox::getMaxCorner(void) {
	return m_maxCorner;
}

inline const Vec3f &BoundingBox::getMaxCorner(void) const {
	return m_maxCorner;
}

inline bool BoundingBox::isInside(const Vec3f &point) const {
	return point[0] > m_minCorner[0] && point[0] < m_maxCorner[0] &&
		   point[1] > m_minCorner[1] && point[1] < m_maxCorner[1] &&
		   point[2] > m_minCorner[2] && point[2] < m_maxCorner[2];
}

#define FINDMINMAX(x0,x1,x2,min,max) \
  min = max = x0;   \
  if(x1<min) min=x1;\
  if(x1>max) max=x1;\
  if(x2<min) min=x2;\
  if(x2>max) max=x2;

/*======================== X-tests ========================*/
#define AXISTEST_X01(a, b, fa, fb)			   \
	p0 = a*v0[1] - b*v0[2];			       	   \
	p2 = a*v2[1] - b*v2[2];			       	   \
        if(p0<p2) {min=p0; max=p2;} else {min=p2; max=p0;} \
	rad = fa * boxhalfsize[1] + fb * boxhalfsize[2];   \
	if(min>rad || max<-rad) return 0;

#define AXISTEST_X2(a, b, fa, fb)			   \
	p0 = a*v0[1] - b*v0[2];			           \
	p1 = a*v1[1] - b*v1[2];			       	   \
        if(p0<p1) {min=p0; max=p1;} else {min=p1; max=p0;} \
	rad = fa * boxhalfsize[1] + fb * boxhalfsize[2];   \
	if(min>rad || max<-rad) return 0;

/*======================== Y-tests ========================*/
#define AXISTEST_Y02(a, b, fa, fb)			   \
	p0 = -a*v0[0] + b*v0[2];		      	   \
	p2 = -a*v2[0] + b*v2[2];	       	       	   \
        if(p0<p2) {min=p0; max=p2;} else {min=p2; max=p0;} \
	rad = fa * boxhalfsize[0] + fb * boxhalfsize[2];   \
	if(min>rad || max<-rad) return 0;

#define AXISTEST_Y1(a, b, fa, fb)			   \
	p0 = -a*v0[0] + b*v0[2];		      	   \
	p1 = -a*v1[0] + b*v1[2];	     	       	   \
        if(p0<p1) {min=p0; max=p1;} else {min=p1; max=p0;} \
	rad = fa * boxhalfsize[0] + fb * boxhalfsize[2];   \
	if(min>rad || max<-rad) return 0;

/*======================== Z-tests ========================*/

#define AXISTEST_Z12(a, b, fa, fb)			   \
	p1 = a*v1[0] - b*v1[1];			           \
	p2 = a*v2[0] - b*v2[1];			       	   \
        if(p2<p1) {min=p2; max=p1;} else {min=p1; max=p2;} \
	rad = fa * boxhalfsize[0] + fb * boxhalfsize[1];   \
	if(min>rad || max<-rad) return 0;

#define AXISTEST_Z0(a, b, fa, fb)			   \
	p0 = a*v0[0] - b*v0[1];				   \
	p1 = a*v1[0] - b*v1[1];			           \
        if(p0<p1) {min=p0; max=p1;} else {min=p1; max=p0;} \
	rad = fa * boxhalfsize[0] + fb * boxhalfsize[1];   \
	if(min>rad || max<-rad) return 0;

inline Vec3f BoundingBox::centerPoint(void) const {
	return (m_maxCorner + m_minCorner)*0.5f;
}

inline Vec3f BoundingBox::halfVector(void) const {
	return (m_maxCorner - m_minCorner)*0.5f;
}

inline Vec3f BoundingBox::size(void) const {
	return (m_maxCorner - m_minCorner);
}

inline void BoundingBox::grow(float minSize, float percentGrow) 
{
	Vec3f bbExpand;
	for(int i=0; i<3; i++)
		bbExpand[i] = ((getMaxCorner()[i] - getMinCorner()[i]) < minSize) ? minSize : 0.0f;

	const Vec3f cp = centerPoint();
	const Vec3f extend = halfVector() * percentGrow;

	setMaxCorner(cp + bbExpand + extend);
	setMinCorner(cp - bbExpand - extend);
}





#endif
