#include "collisiondetection.h"
#include "l3ds/l3dsrenderer.h"

BV* CollDet::createHierarchy(Types type, L3DSRenderer& obj)
{
	std::vector<BV::triangle>* triangles = getTriangles(obj);
	return createHierarchy(type, *triangles);
	delete triangles;
}

BV* CollDet::createHierarchy(Types type, std::vector<BV::triangle>& triangles)
{
	BV*			bvol;
	BSphere*	pBSphere = NULL;
	AABB*		pAABB = NULL;

	if (type==T_AABB) {
		pAABB = new AABB;
		bvol = pAABB;
	}
	if (type==T_BSPHERE) {
		pBSphere = new BSphere;
		bvol = pBSphere;
	}

	// Use top-down approach
	// insert all triangles into one BV

	// Create minimal BV surrounding all triangles
	// Use bounding boxes since it it hard to use other volumes for top-down approach
	BoundingBox bbox0;
	for(unsigned int i=0; i<triangles.size(); i++)
	{
		// Actually, only the vertex list is necessary, but using the triangles is alot easier and likely almost as fast.
		bbox0 += triangles[i].v0;
		bbox0 += triangles[i].v1;
		bbox0 += triangles[i].v2;
	}
	
	if (type==T_AABB) 
		pAABB->bbox = bbox0;
	if (type==T_BSPHERE) {
		pBSphere->pos = bbox0.centerPoint();
		pBSphere->radius = bbox0.halfVector().length();
	}
	
	// Split into 2 boxes or break recursion by storing triangles in this leaf.
	BoundingBox bbox1, bbox2;
	if( (triangles.size() <= 3) || !splitBox(bbox0, bbox1, bbox2, triangles))
	{
		// break recursion
		// insert all triangles into this BV's triangle list
		for(unsigned int i=0; i<triangles.size(); i++)
			bvol->addTriangle(triangles[i]);
		//printf("%d ", triangles.size());
		triangles.clear();
	}
	else
	{
		// splitting into 2 parts along longest axis does not work for a box. Both children will get bounding boxes identical to the parent's.
		// Find longest axis.
//		Vec3f dir;
//		Vec3f dim = bbox0.halfVector();
//		if(dim[0] > dim[1])
//		{
//			if (dim[0] > dim[2]) 
//				dir = Vec3f(dim[0],0,0);
//			else
//				dir = Vec3f(0,0,dim[2]);
//		}
//		else 
//		{
//			if (dim[1] > dim[2]) 
//				dir = Vec3f(0,dim[1],0);
//			else
//				dir = Vec3f(0,0,dim[2]);
//		}
//		
//		BoundingBox bbox1(bbox0.getMinCorner(), bbox0.getMaxCorner()-dir);
//		BoundingBox bbox2(bbox0.getMinCorner()+(dir*0.9999), bbox0.getMaxCorner()); // extend slightly for robustness
		
		// Create 2 triangle list and assign triangles acording to their center point
		std::vector<BV::triangle> triangles1;
		std::vector<BV::triangle> triangles2;
		
		for(unsigned int i=0; i<triangles.size(); i++)
		{
			Vec3f c = (triangles[i].v0 + triangles[i].v1 + triangles[i].v2) * 0.33333333333333f;
			if (bbox1.isInside(c)) 
				triangles1.push_back(triangles[i]);
			else {
				//assert(bbox2.isInside(c));
				triangles2.push_back(triangles[i]);
			}
		}
		triangles.clear(); // to save memory during recursion
		
		// Create 2 children
		BV* child1 =  createHierarchy(type, triangles1);
		BV* child2 =  createHierarchy(type, triangles2);
		
		bvol->addChild(child1);
		bvol->addChild(child2);
	}

	return bvol;
}

bool CollDet::splitBox(BoundingBox& bbox0, BoundingBox& bbox1, BoundingBox& bbox2, std::vector<BV::triangle>& triangles)
{
	// test which of the three axes x,y,z that gives the smallest BV's for the two children

	Vec3f dim = bbox0.halfVector();
	Vec3f axes[] = {Vec3f(dim[0],0,0), Vec3f(0,dim[1],0), Vec3f(0,0,dim[2])};
	int bestAxis = -1;
	float minVolume = FLT_MAX;
	BoundingBox minBox1, minBox2;

	dim = bbox0.size();
	float parentVolume =  dim[0]*dim[1]*dim[2];
	
	
	for(int i=0; i<3; i++)
	{
		Vec3f dir = axes[i];
		dir *= 1.1f;
		BoundingBox bbox1(bbox0.getMinCorner(), bbox0.getMaxCorner()-dir);
		BoundingBox bbox2(bbox0.getMinCorner()+(dir*0.9999f), bbox0.getMaxCorner()); // extend slightly for robustness
		BoundingBox minimalBox1;
		BoundingBox minimalBox2;

		// Asign triangles acording to their center point
		bool bOnly1=true, bOnly2 = true;
		for(unsigned int j=0; j<triangles.size(); j++)
		{
			Vec3f c = (triangles[j].v0 + triangles[j].v1 + triangles[j].v2) * 0.33333333333333f;
			if (bbox1.isInside(c)) 
			{
				minimalBox1 += triangles[j].v0;
				minimalBox1 += triangles[j].v1;
				minimalBox1 += triangles[j].v2;
				bOnly2 = false;
			}
			else {
				//assert(bbox2.isInside(c));
				minimalBox2 += triangles[j].v0;
				minimalBox2 += triangles[j].v1;
				minimalBox2 += triangles[j].v2;
				bOnly1 = false;
			}
		}
		if(bOnly1 || bOnly2)
			continue; // all triangles ended up in just one of the boxes, which is not allowed.

		Vec3f dim = minimalBox1.size();
		float vol1 =  dim[0]*dim[1]*dim[2];
		
		dim = minimalBox2.size();
		float vol2 =  dim[0]*dim[1]*dim[2];
		
//		float totVolume = vol1 + vol2;
		float largestVol = max(vol1, vol2); // find the axis that minimizes the largest of the 2 BVs.

		// we cannot allow any of the created volumes to be identical of parent's volume
		if((vol1<parentVolume) && (vol2<parentVolume) && largestVol < minVolume)
		{
			minVolume = largestVol;
			bestAxis = i;
			minBox1 = bbox1;
			minBox2 = bbox2;
		}
	}
	
	bbox1 = minBox1;
	bbox2 = minBox2;

	return bestAxis != -1;
}

bool CollDet::intersect(BV& obj1, BV& obj2, Mtx4f& mtx)
{
	return BVHierarchy::intersect(obj1, obj2, mtx);
}

void CollDet::findAllIntersections(BV& obj1, BV& obj2, Mtx4f& mtx, std::vector<BV::triangle>& tris)
{
	BVHierarchy::findAllIntersections(obj1, obj2, mtx, tris);
}

void CollDet::findDeepestIntersection(BV& obj1, BV& obj2, Mtx4f& mtx, float& depth)
{
	BVHierarchy::findDeepestIntersection(obj1, obj2, mtx, depth);
}

std::vector<BV::triangle>* CollDet::getTriangles(L3DSRenderer& obj)
{
	std::vector<BV::triangle>* triangles = new std::vector<BV::triangle>;

	// loop through all meshes in the scene
	int totNumTris=0;
	
	for (unsigned int i = 0; i < obj.GetMeshCount(); ++i)
	{
		LMesh &mesh = obj.GetMesh(i);
		totNumTris += mesh.GetTriangleCount();
		
		
		// loop through all triangles
		for(unsigned int j=0; j<mesh.GetTriangleCount(); j++ )
		{
			struct LTriangle2 &tri = mesh.GetTriangle2(j);
			
			Vec3f v0,v1,v2;
			v0 = Vec3f(tri.vertices[0].x, tri.vertices[0].y, tri.vertices[0].z); 
			v1 = Vec3f(tri.vertices[1].x, tri.vertices[1].y, tri.vertices[1].z); 
			v2 = Vec3f(tri.vertices[2].x, tri.vertices[2].y, tri.vertices[2].z); 

			triangles->push_back(BV::triangle(v0, v1, v2));
		}
	}
	return triangles;
}