#ifndef COLLISION_DETECTION_H
#define COLLISION_DETECTION_H

//-----------------------------------------
//
// Code created by Ulf Assarsson, 2006
//
//-----------------------------------------

//#include "l3ds/l3dsrenderer.h"
#include "vecmath.h"
#include "boundingbox.h"
#include <vector>
class L3DSRenderer;

// NOTE: This code may be nicer using templates. However, templates have their problems as well.
// Templates would avoid the dynamic_casts and runtime type info. 

#ifndef max
#define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

int NoDivTriTriIsect(float V0[3],float V1[3],float V2[3], float U0[3],float U1[3],float U2[3]);
					 
class BV
{
public:
	BV() {};
	virtual ~BV() { for(unsigned int i=0; i<m_children.size(); i++) delete m_children[i]; };
	virtual bool intersect(BV& obj2, Mtx4f& mtx) = 0;
	virtual bool deepestIntersect(BV& obj2, Mtx4f& mtx, float depth) = 0;
	bool isLeaf() { return m_children.size() == 0; }

	virtual float volume() = 0;
	
//	virtual void operator+=(const Vec3f& p) = 0; 
	
	int nChildren() { return m_children.size(); }
	BV& child(int index) { return *m_children[index]; }
	void addChild(BV* child) { m_children.push_back(child); }

	bool intersectTriangles(BV& obj2, Mtx4f& mtx)
	{
		for(unsigned int i=0; i<m_triangles.size(); i++) {
			for(unsigned int j=0; j<obj2.m_triangles.size(); j++) {
				if (NoDivTriTriIsect(m_triangles[i].v0.vec, m_triangles[i].v1.vec, m_triangles[i].v2.vec,
					(mtx.multPnt(obj2.m_triangles[j].v0)).vec, 
					(mtx.multPnt(obj2.m_triangles[j].v1)).vec,
					(mtx.multPnt(obj2.m_triangles[j].v2)).vec)) {
					return true;
				}
			}
		}
		return false;
	}

	class triangle
	{
	public:
		triangle() {};
		triangle(Vec3f a, Vec3f b, Vec3f c) { v0=a; v1=b; v2=c;}
		Vec3f v0, v1, v2;
	};
	
	void intersectAllTriangles(BV& obj2, Mtx4f& mtx, std::vector<triangle>&	tris)
	{
		for(unsigned int i=0; i<m_triangles.size(); i++) {
			for(unsigned int j=0; j<obj2.m_triangles.size(); j++) {
				if (NoDivTriTriIsect(m_triangles[i].v0.vec, m_triangles[i].v1.vec, m_triangles[i].v2.vec,
					(mtx.multPnt(obj2.m_triangles[j].v0)).vec, 
					(mtx.multPnt(obj2.m_triangles[j].v1)).vec,
					(mtx.multPnt(obj2.m_triangles[j].v2)).vec)) {
					tris.push_back(m_triangles[i]);
					tris.push_back(obj2.m_triangles[i]);
				}
			}
		}
	}
	
	void findDeepestTriangleIntersection(BV& obj2, Mtx4f& mtx, float&	depth)
	{
		for(unsigned int i=0; i<m_triangles.size(); i++) {
			for(unsigned int j=0; j<obj2.m_triangles.size(); j++) {
				if (NoDivTriTriIsect(m_triangles[i].v0.vec, m_triangles[i].v1.vec, m_triangles[i].v2.vec,
					(mtx.multPnt(obj2.m_triangles[j].v0)).vec, 
					(mtx.multPnt(obj2.m_triangles[j].v1)).vec,
					(mtx.multPnt(obj2.m_triangles[j].v2)).vec)) {
					depth = max(depth, 0.01f); // we should compute depth here
				}
			}
		}
	}
	
	void addTriangle(triangle& tri) { m_triangles.push_back(tri); }
	
private:
	std::vector<BV*>		m_children;
	std::vector<triangle>	m_triangles;
};

class BVHierarchy
{
public:

	static bool intersect(BV& obj1, BV& obj2, Mtx4f& mtx)
	{
		if(!obj1.intersect(obj2, mtx))
			return false;
		if(obj1.isLeaf() && obj2.isLeaf())
			return obj1.intersectTriangles(obj2, mtx);
		else if(!obj1.isLeaf() && !obj2.isLeaf()) {
			if(obj1.volume() > obj2.volume())
				for(int i=0; i<obj1.nChildren(); i++) {
					if(intersect(obj1.child(i), obj2, mtx))
						return true;
				}
				else {
					for(int i=0; i<obj2.nChildren(); i++) {
						if(intersect(obj1, obj2.child(i), mtx))
							return true;
					}
				}
		} 
		else if(obj1.isLeaf() && !obj2.isLeaf()) {
			for(int i=0; i<obj2.nChildren(); i++) {
				if(intersect(obj1, obj2.child(i), mtx))
					return true;
			}
		}
		else {
			for(int i=0; i<obj1.nChildren(); i++) {
				if(intersect(obj1.child(i), obj2, mtx))
					return true;
			}
		}
		return false;
	}	


	static void findAllIntersections(BV& obj1, BV& obj2, Mtx4f& mtx, std::vector<BV::triangle>&	tris)
	{
		if(!obj1.intersect(obj2, mtx))
			return;
		if(obj1.isLeaf() && obj2.isLeaf())
			obj1.intersectAllTriangles(obj2, mtx, tris);
		else if(!obj1.isLeaf() && !obj2.isLeaf()) {
			if(obj1.volume() > obj2.volume())
				for(int i=0; i<obj1.nChildren(); i++) {
					findAllIntersections(obj1.child(i), obj2, mtx, tris);
				}
				else {
					for(int i=0; i<obj2.nChildren(); i++) {
						findAllIntersections(obj1, obj2.child(i), mtx, tris);
					}
				}
		} 
		else if(obj1.isLeaf() && !obj2.isLeaf()) {
			for(int i=0; i<obj2.nChildren(); i++) {
				findAllIntersections(obj1, obj2.child(i), mtx, tris);
			}
		}
		else {
			for(int i=0; i<obj1.nChildren(); i++) {
				findAllIntersections(obj1.child(i), obj2, mtx, tris);
			}
		}
	}
	
	static void findDeepestIntersection(BV& obj1, BV& obj2, Mtx4f& mtx, float&	depth)
	{
		if(!obj1.deepestIntersect(obj2, mtx, depth))
			return;
		if(obj1.isLeaf() && obj2.isLeaf())
			obj1.findDeepestTriangleIntersection(obj2, mtx, depth);
		else if(!obj1.isLeaf() && !obj2.isLeaf()) {
			if(obj1.volume() > obj2.volume())
				for(int i=0; i<obj1.nChildren(); i++) {
					findDeepestIntersection(obj1.child(i), obj2, mtx, depth);
				}
				else {
					for(int i=0; i<obj2.nChildren(); i++) {
						findDeepestIntersection(obj1, obj2.child(i), mtx, depth);
					}
				}
		} 
		else if(obj1.isLeaf() && !obj2.isLeaf()) {
			for(int i=0; i<obj2.nChildren(); i++) {
				findDeepestIntersection(obj1, obj2.child(i), mtx, depth);
			}
		}
		else {
			for(int i=0; i<obj1.nChildren(); i++) {
				findDeepestIntersection(obj1.child(i), obj2, mtx, depth);
			}
		}
	}	
	
	
private:
	
};


class AABB : public virtual BV
{
public:
	virtual bool intersect(BV& obj2, Mtx4f& mtx)
	{
		AABB* o2 = dynamic_cast<AABB*>(&obj2);
		if (o2) 
			return intersect(*o2, mtx);
		return false;
	}
	bool intersect(AABB& obj2, Mtx4f& mtx)
	{
		// create AABB around transformed obj2.
		BoundingBox box2;
		Vec3f minpt = obj2.bbox.getMinCorner();
		Vec3f dim = obj2.bbox.getMaxCorner() - obj2.bbox.getMinCorner();
		box2 += mtx.multPnt(minpt);
		box2 += mtx.multPnt(minpt+Vec3f(0,0,dim[2]));
		box2 += mtx.multPnt(minpt+Vec3f(0,dim[1],0));
		box2 += mtx.multPnt(minpt+Vec3f(dim[0],0,0));
		box2 += mtx.multPnt(minpt+Vec3f(dim[0],dim[1],0));
		box2 += mtx.multPnt(minpt+Vec3f(0,dim[1],dim[2]));
		box2 += mtx.multPnt(minpt+Vec3f(dim[0],0,dim[2]));
		box2 += mtx.multPnt(minpt+Vec3f(dim[0],dim[1],dim[2]));
		
		Vec3f dim1 = bbox.getMaxCorner() - bbox.getMinCorner();
		Vec3f dim2 = box2.getMaxCorner() - box2.getMinCorner();

		for(int i=0; i<3; i++)
		{
			if((box2.getMaxCorner()[i] < bbox.getMinCorner()[i]) ||
			   (bbox.getMaxCorner()[i] < box2.getMinCorner()[i]))
			   return false;
		}
		return true;
	};
	
	virtual bool deepestIntersect(BV& obj2, Mtx4f& mtx, float depth)
	{
		AABB* o2 = dynamic_cast<AABB*>(&obj2);
		if (o2) 
			return deepestIntersect(*o2, mtx, depth);
		return false;
	}
	
	bool deepestIntersect(AABB& obj2, Mtx4f& mtx, float depth)
	{
		// Find size of overlapping box-shaped region between the 2 boxes.
		// Compute the length of the diagonal return true if it is larger than depth. Else false.
		// Create AABB around transformed obj2.
		BoundingBox box2;
		Vec3f minpt = obj2.bbox.getMinCorner();
		Vec3f dim = obj2.bbox.getMaxCorner() - obj2.bbox.getMinCorner();
		box2 += mtx.multPnt(minpt);
		box2 += mtx.multPnt(minpt+Vec3f(0,0,dim[2]));
		box2 += mtx.multPnt(minpt+Vec3f(0,dim[1],0));
		box2 += mtx.multPnt(minpt+Vec3f(dim[0],0,0));
		box2 += mtx.multPnt(minpt+Vec3f(dim[0],dim[1],0));
		box2 += mtx.multPnt(minpt+Vec3f(0,dim[1],dim[2]));
		box2 += mtx.multPnt(minpt+Vec3f(dim[0],0,dim[2]));
		box2 += mtx.multPnt(minpt+Vec3f(dim[0],dim[1],dim[2]));
		
		Vec3f dim1 = bbox.getMaxCorner() - bbox.getMinCorner();
		Vec3f dim2 = box2.getMaxCorner() - box2.getMinCorner();
		Vec3f dimReg;
		for(int i=0; i<3; i++)
		{
			if((box2.getMaxCorner()[i] < bbox.getMinCorner()[i]) ||
				(bbox.getMaxCorner()[i] < box2.getMinCorner()[i]))
				return false;
			dimReg[i] = max( (box2.getMaxCorner()[i] - bbox.getMinCorner()[i]),
							 (bbox.getMaxCorner()[i] - box2.getMinCorner()[i]));
			assert(dimReg[i] >= 0);
		}
		return (dimReg.length() > depth);
	};
	
	virtual float volume() 
	{
		Vec3f dim = bbox.size();
		return dim[0]*dim[1]*dim[2];
	}

//	virtual void operator+=(const Vec3f& p) { bbox += p; } 
		
	BoundingBox bbox;
};

class BSphere : public virtual BV
{
public:
	BSphere() : radius(-1) {}

	virtual bool intersect(BV& obj2, Mtx4f& mtx)
	{
		BSphere* o2 = dynamic_cast<BSphere*>(&obj2);
		if (o2) 
			return intersect(*o2, mtx);
		return false;
	}
	bool intersect(BSphere& obj2, Mtx4f& mtx) 
	{
		// assume no scaling
		Vec3f center2 = mtx.multPnt(obj2.pos);
		float r = radius + obj2.radius;
		return ( (center2 - pos).length2() < r*r );
	}

	virtual bool deepestIntersect(BV& obj2, Mtx4f& mtx, float depth)
	{
		BSphere* o2 = dynamic_cast<BSphere*>(&obj2);
		if (o2) 
			return deepestIntersect(*o2, mtx, depth);
		return false;
	}
	bool deepestIntersect(BSphere& obj2, Mtx4f& mtx, float depth) 
	{
		// spheres must intersect at least with depth
		// assume no scaling
		Vec3f center2 = mtx.multPnt(obj2.pos);
		float r = max(0, radius + obj2.radius - depth);
		return ( (center2 - pos).length2() < r*r );
	}
	
	
	virtual float volume() 
	{
		return 4*3.14159265358979323846f*radius*radius*radius*0.33333f;
	}

//	virtual void operator+=(const Vec3f& p) { bbox += p; pos = bbox.centerPoint(); radius = bbox.halfVector().length(); } 
	
	float radius;
	Vec3f pos;
//	BoundingBox bbox; // this is for construction of the bsphere. Hard to create from consequtive points otherwise.
};


class CollDet
{
public:
	CollDet() {};
	~CollDet() {};

	enum Types { T_AABB, T_BSPHERE};
	
	static BV* createHierarchy(Types type, L3DSRenderer& obj);
	static BV* createHierarchy(Types type, std::vector<BV::triangle>& triangles);
		
	static bool intersect(BV& obj1, BV& obj2, Mtx4f& mtx); // mtx transforms from obj2-space into obj1-space. I.e., mtx = (obj1.mtx)^-1 * (obj2.mtx)
	static void findAllIntersections(BV& obj1, BV& obj2, Mtx4f& mtx, std::vector<BV::triangle>&	tris);
	static void findDeepestIntersection(BV& obj1, BV& obj2, Mtx4f& mtx, float&	depth);
	
private:
	static std::vector<BV::triangle>* getTriangles(L3DSRenderer& obj);
	static bool splitBox(BoundingBox& bbox0, BoundingBox& bbox1, BoundingBox& bbox2, std::vector<BV::triangle>& triangles);
		
		
};
#endif //COLLISION_DETECTION_H