#pragma warning (disable:4786)

#include <windows.h>
#include "missile.h"
#include <GL/GLee.h>
#include <gl/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include "tex/texturemanager.h"
#include "nature.h"
#include "Sound/bassound.h"


#ifndef M_PI
#define M_PI       3.14159265358979323846
#endif

double		 	 Missile::m_maxLifeTime = 2.5;	// A missile only has this short life time, in seconds
std::vector<int> Missile::m_explTexIDs;			// Explosion texture IDs

L3DSRenderer Missile::m_missileModel;		


Missile::Missile() : Object() // This means that the Missile constructor calls the default Object constructor
{
	static bool bFirstTime=true; // static means that this object is the very same for all invocations of this method
	if(bFirstTime)
	{
		bFirstTime=false;
		m_missileModel.Load("media/models", "sidewinder.3ds"); // Loads 3ds-model
		Mtx4f m0, m1, m2, m3;
		m0.loadIdentity();
		m1.loadIdentity();
		m2.loadIdentity();
		m3.loadIdentity();
		m1.scale(5, 5, 5);
		m2.rotX(-M_PI*0.5f);
		//m3.rotY(-M_PI*1.0f);
		Vec3f center = m_missileModel.GetBBox().centerPoint();
		m0.translate(-center[0], -center[1], -center[2]);
		m_missileModel.getMatrix() = m3*m2*m1*m0; // translate to center, scale by 0.01 and rotate -90 degrees around x-axis
	}

	m_l3dsObj = &m_missileModel;
	m_useBackFaceCulling = false;
}

Missile::~Missile()
{

}

void Missile::update(double t, CNature* nature, CBassSound* explSound)
{
	if(m_timeLastUpdate == 0)
	{
		m_timeOfBirth = t;
		m_currentTexID = m_baseTexID;
		m_timeLastUpdate = t;
	}

	// update position
	if (m_status == Object::NORMAL) 
	{
		if(t - m_timeOfBirth >= m_maxLifeTime)
		{
			m_status = DEAD;
		}
		else 
		{
			Mtx4f transl;
			Vec3f v = m_speedVec * ((t-m_timeLastUpdate)*50.0f);
			transl.translate(v[0], v[1], v[2]);
			m_mtx = transl * m_mtx;

			// check collision with the ground
			Vec3f pos, lookVec, upVec;
			getPositionAndOrientation(pos, lookVec, upVec);
			Vec3f front = pos + lookVec*getRadius();
			Vec3f back  = pos - lookVec*getRadius();
			float h0 = nature->getHeight(front[0], front[2]);
			float h1 = nature->getHeight(back[0], back[2]);
			if(front[1]<h0 || back[1]<h1)
			{
				m_status = Object::EXPLODING;
				explSound->play();
			}
		}
	}
	
	if ( m_status == Object::EXPLODING)
	{
		if(m_timeOfExplosion == 0)
		{
			m_timeOfExplosion = t;
			initParticles();
		}
		
		float ix = (t-m_timeOfExplosion)*m_explTexIDs.size(); // move through whole animation in 1s
		if(ix >= m_explTexIDs.size())
		{
			m_status = DEAD;
			m_particles.clear();
		}
		else
		{
			int index = ix;
			m_currentTexID = m_explTexIDs[index];
		}
	}

	m_timeLastUpdate = t;
}

void Missile::loadExplosionTextures()
{
	for(int i=1; i<=10; i++)
	{
		char buf[60];
		sprintf(buf,"media/textures/plazmablow/plazmablowpage_%2.2d.ppm",i);

		unsigned char transp[] = {0,0,0};
		m_explTexIDs.push_back(TextureManager::load(buf, transp, true));
		
	}
}

void Missile::render(Camera cam)
{
	if (m_status == Object::NORMAL) 
	{
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glMultMatrixf(m_mtx.array);
		glMultMatrixf(m_l3dsObj->getMatrix().array);
		
		m_l3dsObj->RenderOGL(m_useBackFaceCulling);
		
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}
	else if(m_status == Object::EXPLODING) 
	{
		glPushAttrib(GL_ALL_ATTRIB_BITS); // save current OpenGL state
		
		// Draw a billboard displaying the explosion texture animation
		//	glActiveTexture(GL_TEXTURE0 + 0); 
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, m_currentTexID);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		
		glDisable(GL_CULL_FACE);
		//glDepthMask(0); // disable updating of z-buffer
		glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE, GL_REPLACE); //GL_MODULATE
		glEnable(GL_BLEND);
		glColor4f(1,1,1,1);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//		glEnable(GL_ALPHA_TEST);
//		glAlphaFunc(GL_GREATER, 0.1);
		
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		
		Vec3f c = getCenterPoint();
		float r = 2*getRadius();
		
		// Create a quad in screenX and screenY directions
		Vec3f lookVec = cam.m_lookAt - cam.m_lookFrom;
		lookVec.normalize();
		
		Vec3f screenXaxis = lookVec.cross(cam.m_upVec);
		screenXaxis.normalize();
		
		Vec3f up = screenXaxis.cross(lookVec);
		up.normalize();
		
		Vec3f v0 = c - up*0.5f*r - screenXaxis*0.5f*r;
		Vec3f v1 = v0 + screenXaxis*r;
		Vec3f v2 = v1 + up*r;
		Vec3f v3 = v0 + up*r;
		
		
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3fv(v0.vec);
		
		glTexCoord2f(1,0);
		glVertex3fv(v1.vec);
		
		glTexCoord2f(1,1);
		glVertex3fv(v2.vec);
		
		glTexCoord2f(0,1);
		glVertex3fv(v3.vec);
		glEnd();
		
		glPopMatrix();
		
		glDisable(GL_BLEND);		// These 3 lines are actually not necessary since we restore full OpenGL-state below
		glDisable(GL_TEXTURE_2D);	
		glDepthMask(1);				// enable updating of z-buffer
		
		glPopAttrib();				// restore OpenGL-state



		// ---ADD EXPLOSION PARTICLE SYSTEM HERE: ---
		glPushAttrib(GL_ALL_ATTRIB_BITS);

		float t = m_timeLastUpdate-m_timeOfExplosion;

		// UPDATE YOUR PARTICLE POSITIONS AND RENDER THEM USING E.G GL_POINTS
		// ADD YOUR CODE HERE:

		// This line shows you how to loop over the std::vector of particles
		for(unsigned int i=0; i<m_particles.size(); i++)
		{
			Vec3f pos	= m_particles[i].m_pos;			// shows how to get the initial position of the particle
			Vec3f speed = m_particles[i].m_speedVec;	// shows how to get the speed vector of the particle

		}

		glPopAttrib();
		// --------------------------------
	}
}
void Missile::initParticles()
{
	// Initiate your particles here. 
	// E.g. initiate 500 particles at center of missile 
	// and set their speed vectors to a random unit vector 
	// random value in [0,1] can be generated by:
	// float a = (float)rand()/(float)RAND_MAX;  

	Vec3f c = getCenterPoint();			// center of missile in world space

	for(unsigned int i=0; i<500; i++)
	{
		Particle p;
		//p.m_pos = ...
		//p.m_speedVec = ...


		m_particles.push_back(p);
	}
}

