#pragma warning (disable:4786)

#include <windows.h>
#include "object.h"
#include <GL/GLee.h>
#include <gl/glut.h>
#include <stdio.h>
#include <stdlib.h>

#include "tex/texturemanager.h"

std::vector<int> Object::m_explTexIDs;			// Explosion texture IDs

Object::Object() : m_status(NORMAL), m_killerID(-1)
{
	// constructor. Do allocations per object here
	m_timeLastUpdate = 0;
	m_timeOfExplosion = 0;

	m_useBackFaceCulling = true;

	m_mtx.loadIdentity();
}

Object::~Object()
{
	// destructor. Do deallocations per object here
}

float Object::getRadius()
{
	// Returns the physical radius of the object in world space.
	// This method is not fool proof for objects with asymetric scaling (in x,y,z) 
	Vec3f hv = m_l3dsObj->GetBBox().halfVector();

	hv = m_l3dsObj->getMatrix().multVec(hv);
	float r = hv.length();
	return r;
}

Vec3f Object::getCenterPoint()
{
	// returns the center point of the object
	Vec3f c = m_l3dsObj->GetBBox().centerPoint();
	
	Mtx4f m = m_mtx * m_l3dsObj->getMatrix();
	c = m.multPnt(c);
	return c;
}

BBox3f Object::getBoundingBoxInWorldSpace()
{
	// returns axis aligned bounding box in world coordinate system
	Vec3f c = m_l3dsObj->GetBBox().centerPoint();
	Mtx4f mtx = m_mtx * m_l3dsObj->getMatrix();
	
	// Transform obj2 with mtx. I.e., create a bbox around the 8 transformed corner points.
	// This can be optimized by only transforming origin + the 3 axes instead.
	BBox3f transformedBox2;
	Vec3f minpt = m_l3dsObj->GetBBox().getMinCorner();
	Vec3f dim   = m_l3dsObj->GetBBox().getMaxCorner() - minpt;
	transformedBox2 += mtx.multPnt(minpt);
	transformedBox2 += mtx.multPnt(minpt+Vec3f(0,0,dim[2]));
	transformedBox2 += mtx.multPnt(minpt+Vec3f(0,dim[1],0));
	transformedBox2 += mtx.multPnt(minpt+Vec3f(dim[0],0,0));
	transformedBox2 += mtx.multPnt(minpt+Vec3f(dim[0],dim[1],0));
	transformedBox2 += mtx.multPnt(minpt+Vec3f(0,dim[1],dim[2]));
	transformedBox2 += mtx.multPnt(minpt+Vec3f(dim[0],0,dim[2]));
	transformedBox2 += mtx.multPnt(minpt+Vec3f(dim[0],dim[1],dim[2]));

	return transformedBox2;
}

bool Object::isColliding(Object& obj2) // checks for collision with another object
{
	// transform from obj2-space into obj-space.
	// I.e., go from obj2-space into world space into obj-space.
	// I.e., m = obj^-1 * obj2
	Mtx4f m = m_mtx * m_l3dsObj->getMatrix();
	m.invert();
	m = m * (obj2.m_mtx * obj2.m_l3dsObj->getMatrix());

	bool bHit;
	if(true)
	{
		// This tests exact collision
		bHit = CollDet::intersect(*(m_l3dsObj->getBVH()), *(obj2.m_l3dsObj->getBVH()), m);
	}
	else
	{
		// This tests approximate collision. If exact collision is to expensive, use this code path instead
		// If so, offset missile at start a little bit further from the player to guarantee it starts outside the bounding box
		m.loadIdentity();
		bHit = getBoundingBoxInWorldSpace().intersect(obj2.getBoundingBoxInWorldSpace(), m);
	}
	
	return bHit;
}

void Object::loadExplosionTextures()
{
	for(int i=1; i<=10; i++)
	{
		char buf[60];
		sprintf(buf,"media/textures/plazmablow/plazmablowpage_%2.2d.ppm",i);

		unsigned char transp[] = {0,0,0};
		m_explTexIDs.push_back(TextureManager::load(buf, transp, true));
		
	}
}

void Object::update(double t)
{
	if (m_status == Object::NORMAL) {
		m_currentTexID = m_baseTexID;
	}
	else if ( m_status == Object::EXPLODING)
	{
		if(m_timeOfExplosion == 0)
			m_timeOfExplosion = t;
		
		float ix = (t-m_timeOfExplosion)*m_explTexIDs.size(); // move through whole animation in 1s
		if(ix > m_explTexIDs.size())
		{
			m_status = DEAD;
			m_timeOfExplosion = 0;
		}
		else
		{
			int index = ix;
			m_currentTexID = m_explTexIDs[index];
		}
	}
	else if (m_status == DEAD)
	{
		m_currentTexID = m_baseTexID;
		m_status = NORMAL;
	}

	m_timeLastUpdate = t;
	
}

void Object::render(Camera cam)
{
	if (m_status == Object::NORMAL) 
	{
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glMultMatrixf(m_mtx.array);
		glMultMatrixf(m_l3dsObj->getMatrix().array);
		
		m_l3dsObj->RenderOGL(m_useBackFaceCulling);
		
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		
	}
	else if(m_status == Object::EXPLODING) 
	{
		glPushAttrib(GL_ALL_ATTRIB_BITS); // save current OpenGL state

		// Draw a billboard displaying the explosion texture animation
		// This billboard is semitransparent
		// Remember that for correctness, we should actually render all semitransparent objects last, in back-to-front order
		// We are lazy and skip that for now and just make sure that the background (floor) is rendered before our exlosions

		//	glActiveTexture(GL_TEXTURE0 + 0); 
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, m_currentTexID);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		
		glDisable(GL_CULL_FACE);
		glDepthMask(0); // disable updating of z-buffer
		glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE, GL_REPLACE); //GL_MODULATE
		glEnable(GL_BLEND);
		glColor4f(1,1,1,1);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//		glEnable(GL_ALPHA_TEST);
//		glAlphaFunc(GL_GREATER, 0.1);
		
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		
		Vec3f c = getCenterPoint();
		float r = 2*getRadius();

		// Create a quad in screenX and screenY directions
		Vec3f lookVec = cam.m_lookAt - cam.m_lookFrom;
		lookVec.normalize();

		Vec3f screenXaxis = lookVec.cross(cam.m_upVec);
		screenXaxis.normalize();

		Vec3f up = screenXaxis.cross(lookVec);
		up.normalize();
		
		Vec3f v0 = c - up*0.5f*r - screenXaxis*0.5f*r;
		Vec3f v1 = v0 + screenXaxis*r;
		Vec3f v2 = v1 + up*r;
		Vec3f v3 = v0 + up*r;
		
		
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3fv(v0.vec);
		
		glTexCoord2f(1,0);
		glVertex3fv(v1.vec);
		
		glTexCoord2f(1,1);
		glVertex3fv(v2.vec);
		
		glTexCoord2f(0,1);
		glVertex3fv(v3.vec);
		glEnd();
		
		glPopMatrix();
		
		glDisable(GL_BLEND);		// These 3 lines are actually not necessary since we restore full OpenGL-state below
		glDisable(GL_TEXTURE_2D);	
		glDepthMask(1);				// enable updating of z-buffer

		glPopAttrib();				// restore OpenGL-state
	}
}

void Object::getPositionAndOrientation(Vec3f& pos, Vec3f& lookVec, Vec3f& upVec)
{
	// Get the position and orientation from m_mtx
	// Define y-axis as up direction and z-axis as forward direction
	// m_l3dsObj->m_transform can be set to rotate the l3ds-object if the model happens to be modelled differently 

	pos[0]		= m_mtx.array[12];	pos[1]		= m_mtx.array[13];	pos[2]		= m_mtx.array[14];
	lookVec[0]	= m_mtx.array[8];	lookVec[1]	= m_mtx.array[9];	lookVec[2]	= m_mtx.array[10];
	upVec[0]	= m_mtx.array[4];	upVec[1]	= m_mtx.array[5];	upVec[2]	= m_mtx.array[6];
	
}

void Object::setPositionAndOrientation(Vec3f pos, Vec3f lookVec, Vec3f upVec)
{
	// Set the position and orientation to m_mtx
//	Vec3f xVec = upVec.cross(lookVec); 
	Vec3f xVec = lookVec.cross(upVec); 
	xVec.normalize(); // it is not really necessary to normalize since upVec and lookVec are normalized.
	
	// Recompute upVec so that it surely is perpendicular to xVec and lookVec
	upVec = xVec.cross(lookVec);
	//upVec = lookVec.cross(xVec);

	// Set to matrix that rotates z-axis to lookVec and y-axis to upVec
	m_mtx.billboardMatrix(Vec3f(0,0,1), Vec3f(0,1,0), lookVec, upVec);
		
	m_mtx.array[12]	= pos[0];		m_mtx.array[13] = pos[1];		m_mtx.array[14] = pos[2];		m_mtx.array[15] = 1.0f;
//	m_mtx.array[8]	= lookVec[0];	m_mtx.array[9]  = lookVec[1];	m_mtx.array[10] = lookVec[2];	m_mtx.array[11] = 0.0f;
//	m_mtx.array[4]	= upVec[0];		m_mtx.array[5]  = upVec[1];		m_mtx.array[6]  = upVec[2];		m_mtx.array[7]  = 0.0f;
//
	return;
	
	m_mtx.mtx[0][0] = xVec[0];
	m_mtx.mtx[1][0] = xVec[1];
	m_mtx.mtx[2][0] = xVec[2];
	m_mtx.mtx[3][0] = pos[0];
	
	m_mtx.mtx[0][1] = upVec[0];
	m_mtx.mtx[1][1] = upVec[1];
	m_mtx.mtx[2][1] = upVec[2];
	m_mtx.mtx[3][1] = pos[1];

	m_mtx.mtx[0][2] = lookVec[0];
	m_mtx.mtx[1][2] = lookVec[1];
	m_mtx.mtx[2][2] = lookVec[2];
	m_mtx.mtx[3][2] = pos[2];

	// 4:th row (bottom row of matrix)
	m_mtx.mtx[0][3] = 0;
	m_mtx.mtx[1][3] = 0;
	m_mtx.mtx[2][3] = 0;
	m_mtx.mtx[3][3] = 1.0f;
	
}


